@@ -1,14 +1,13 @@
 use std::{fs, path::Path};
 
-use ansi_term::Colour;
+use anstyle::AnsiColor;
 use anyhow::{anyhow, Result};
-use tree_sitter::Point;
 use tree_sitter_loader::{Config, Loader};
 use tree_sitter_tags::{TagsConfiguration, TagsContext};
 
 use super::{
-    query_testing::{parse_position_comments, Assertion},
-    test::opt_color,
+    query_testing::{parse_position_comments, to_utf8_point, Assertion, Utf8Point},
+    test::paint,
     util,
 };
 
@@ -50,50 +49,85 @@ pub fn test_tags(
     directory: &Path,
     use_color: bool,
 ) -> Result<()> {
-    let mut failed = false;
     println!("tags:");
+    test_tags_indented(loader, loader_config, tags_context, directory, use_color, 2)
+}
+
+pub fn test_tags_indented(
+    loader: &Loader,
+    loader_config: &Config,
+    tags_context: &mut TagsContext,
+    directory: &Path,
+    use_color: bool,
+    indent_level: usize,
+) -> Result<()> {
+    let mut failed = false;
+
     for tag_test_file in fs::read_dir(directory)? {
         let tag_test_file = tag_test_file?;
         let test_file_path = tag_test_file.path();
         let test_file_name = tag_test_file.file_name();
-        let (language, language_config) = loader
-            .language_configuration_for_file_name(&test_file_path)?
-            .ok_or_else(|| {
-                anyhow!(
-                    "{}",
-                    util::lang_not_found_for_path(test_file_path.as_path(), loader_config)
-                )
-            })?;
-        let tags_config = language_config
-            .tags_config(language)?
-            .ok_or_else(|| anyhow!("No tags config found for {:?}", test_file_path))?;
-        match test_tag(
-            tags_context,
-            tags_config,
-            fs::read(&test_file_path)?.as_slice(),
-        ) {
-            Ok(assertion_count) => {
-                println!(
-                    "  ✓ {} ({assertion_count} assertions)",
-                    opt_color(
-                        use_color,
-                        Colour::Green,
-                        test_file_name.to_string_lossy().as_ref()
-                    ),
-                );
-            }
-            Err(e) => {
-                println!(
-                    "  ✗ {}",
-                    opt_color(
-                        use_color,
-                        Colour::Red,
-                        test_file_name.to_string_lossy().as_ref()
-                    )
-                );
-                println!("    {e}");
+        print!(
+            "{indent:indent_level$}",
+            indent = "",
+            indent_level = indent_level * 2
+        );
+        if test_file_path.is_dir() && test_file_path.read_dir()?.next().is_some() {
+            println!("{}:", test_file_name.to_string_lossy());
+            if test_tags_indented(
+                loader,
+                loader_config,
+                tags_context,
+                &test_file_path,
+                use_color,
+                indent_level + 1,
+            )
+            .is_err()
+            {
                 failed = true;
             }
+        } else {
+            let (language, language_config) = loader
+                .language_configuration_for_file_name(&test_file_path)?
+                .ok_or_else(|| {
+                    anyhow!(
+                        "{}",
+                        util::lang_not_found_for_path(test_file_path.as_path(), loader_config)
+                    )
+                })?;
+            let tags_config = language_config
+                .tags_config(language)?
+                .ok_or_else(|| anyhow!("No tags config found for {test_file_path:?}"))?;
+            match test_tag(
+                tags_context,
+                tags_config,
+                fs::read(&test_file_path)?.as_slice(),
+            ) {
+                Ok(assertion_count) => {
+                    println!(
+                        "✓ {} ({assertion_count} assertions)",
+                        paint(
+                            use_color.then_some(AnsiColor::Green),
+                            test_file_name.to_string_lossy().as_ref()
+                        ),
+                    );
+                }
+                Err(e) => {
+                    println!(
+                        "✗ {}",
+                        paint(
+                            use_color.then_some(AnsiColor::Red),
+                            test_file_name.to_string_lossy().as_ref()
+                        )
+                    );
+                    println!(
+                        "{indent:indent_level$}  {e}",
+                        indent = "",
+                        indent_level = indent_level * 2
+                    );
+                    failed = true;
+                }
+            }
         }
     }
 
@@ -117,11 +151,13 @@ pub fn test_tag(
     let mut actual_tags = Vec::<&String>::new();
     for Assertion {
         position,
+        length,
         negative,
         expected_capture_name: expected_tag,
     } in &assertions
     {
         let mut passed = false;
+        let mut end_column = position.column + length - 1;
 
         'tag_loop: while let Some(tag) = tags.get(i) {
             if tag.1 <= *position {
@@ -133,7 +169,8 @@ pub fn test_tag(
             // position, looking for one that matches the assertion
             let mut j = i;
             while let (false, Some(tag)) = (passed, tags.get(j)) {
-                if tag.0 > *position {
+                end_column = position.column + length - 1;
+                if tag.0.column > end_column {
                     break 'tag_loop;
                 }
 
@@ -155,7 +192,7 @@ pub fn test_tag(
         if !passed {
             return Err(Failure {
                 row: position.row,
-                column: position.column,
+                column: end_column,
                 expected_tag: expected_tag.clone(),
                 actual_tags: actual_tags.into_iter().cloned().collect(),
             }
@@ -170,7 +207,7 @@ pub fn get_tag_positions(
     tags_context: &mut TagsContext,
     tags_config: &TagsConfiguration,
     source: &[u8],
-) -> Result<Vec<(Point, Point, String)>> {
+) -> Result<Vec<(Utf8Point, Utf8Point, String)>> {
     let (tags_iter, _has_error) = tags_context.generate_tags(tags_config, source, None)?;
     let tag_positions = tags_iter
         .filter_map(std::result::Result::ok)
@@ -181,7 +218,11 @@ pub fn get_tag_positions(
             } else {
                 format!("reference.{tag_postfix}")
             };
-            (tag.span.start, tag.span.end, tag_name)
+            (
+                to_utf8_point(tag.span.start, source),
+                to_utf8_point(tag.span.end, source),
+                tag_name,
+            )
         })
         .collect();
     Ok(tag_positions)
