@@ -1,28 +1,27 @@
 use std::{
     env, fs,
     path::{Path, PathBuf},
+    sync::LazyLock,
 };
 
 use anyhow::Context;
-use lazy_static::lazy_static;
 use tree_sitter::Language;
+use tree_sitter_generate::{load_grammar_file, ALLOC_HEADER, ARRAY_HEADER};
 use tree_sitter_highlight::HighlightConfiguration;
 use tree_sitter_loader::{CompileConfig, Loader};
 use tree_sitter_tags::TagsConfiguration;
 
-use crate::generate::ALLOC_HEADER;
+use crate::tests::generate_parser;
 
 include!("./dirs.rs");
 
-lazy_static! {
-    static ref TEST_LOADER: Loader = {
-        let mut loader = Loader::with_parser_lib_path(SCRATCH_DIR.clone());
-        if env::var("TREE_SITTER_GRAMMAR_DEBUG").is_ok() {
-            loader.use_debug_build(true);
-        }
-        loader
-    };
-}
+static TEST_LOADER: LazyLock<Loader> = LazyLock::new(|| {
+    let mut loader = Loader::with_parser_lib_path(SCRATCH_DIR.clone());
+    if env::var("TREE_SITTER_GRAMMAR_DEBUG").is_ok() {
+        loader.debug_build(true);
+    }
+    loader
+});
 
 pub fn test_loader() -> &'static Loader {
     &TEST_LOADER
@@ -43,6 +42,13 @@ pub fn get_language(name: &str) -> Language {
     TEST_LOADER.load_language_at_path(config).unwrap()
 }
 
+pub fn get_test_fixture_language(name: &str) -> Language {
+    let grammar_dir_path = fixtures_dir().join("test_grammars").join(name);
+    let grammar_json = load_grammar_file(&grammar_dir_path.join("grammar.js"), None).unwrap();
+    let (parser_name, parser_code) = generate_parser(&grammar_json).unwrap();
+    get_test_language(&parser_name, &parser_code, Some(&grammar_dir_path))
+}
+
 pub fn get_language_queries_path(language_name: &str) -> PathBuf {
     GRAMMARS_DIR.join(language_name).join("queries")
 }
@@ -85,7 +91,7 @@ pub fn get_test_language(name: &str, parser_code: &str, path: Option<&Path>) ->
     fs::create_dir_all(&src_dir).unwrap();
 
     let parser_path = src_dir.join("parser.c");
-    if !fs::read_to_string(&parser_path).map_or(false, |content| content == parser_code) {
+    if !fs::read_to_string(&parser_path).is_ok_and(|content| content == parser_code) {
         fs::write(&parser_path, parser_code).unwrap();
     }
 
@@ -94,8 +100,7 @@ pub fn get_test_language(name: &str, parser_code: &str, path: Option<&Path>) ->
         if scanner_path.exists() {
             let scanner_code = fs::read_to_string(&scanner_path).unwrap();
             let scanner_copy_path = src_dir.join("scanner.c");
-            if !fs::read_to_string(&scanner_copy_path)
-                .map_or(false, |content| content == scanner_code)
+            if !fs::read_to_string(&scanner_copy_path).is_ok_and(|content| content == scanner_code)
             {
                 fs::write(&scanner_copy_path, scanner_code).unwrap();
             }
@@ -110,18 +115,16 @@ pub fn get_test_language(name: &str, parser_code: &str, path: Option<&Path>) ->
     let header_path = src_dir.join("tree_sitter");
     fs::create_dir_all(&header_path).unwrap();
 
-    [
+    for (file, content) in [
         ("alloc.h", ALLOC_HEADER),
-        ("array.h", tree_sitter::ARRAY_HEADER),
+        ("array.h", ARRAY_HEADER),
         ("parser.h", tree_sitter::PARSER_HEADER),
-    ]
-    .iter()
-    .for_each(|(file, content)| {
+    ] {
         let file = header_path.join(file);
         fs::write(&file, content)
             .with_context(|| format!("Failed to write {:?}", file.file_name().unwrap()))
             .unwrap();
-    });
+    }
 
     let paths_to_check = if let Some(scanner_path) = &scanner_path {
         vec![parser_path, scanner_path.clone()]
